using System;
using System.Configuration;
using System.IO;
using System.Data.SqlClient;
using System.Net;
using System.Reflection;
using System.ServiceProcess;
using System.Text;
using System.Xml;
using ITHit.WebDAV.Server.Logger;
using ITHit.WebDAV.Server;

namespace WebDAVServer_Class1Listener
{
    /// <summary>
    /// WebDAV engine host.
    /// </summary>
    internal class Program
    {
        public static bool Listening { get; set; }

        private static DavEngine engine;

        /// <summary>
        /// Whether requests/responses shall be logged.
        /// </summary>
        private static readonly bool debugLoggingEnabled =
            ConfigurationManager.AppSettings["DebugLoggingEnabled"].Equals(
                "true",
                StringComparison.InvariantCultureIgnoreCase);

        /// <summary>
        /// Logger instance.
        /// </summary>
        private static readonly DefaultLoggerImpl logger = new DefaultLoggerImpl();
        /// <summary>
        /// Gets a value indicating whether the program is run as Windows service or standalone application.
        /// </summary>
        private static bool IsServiceMode
        {
            get { return !Environment.UserInteractive; }
        }

        /// <summary>
        /// Entry point.
        /// </summary>
        /// <param name="args">Command line arguments.</param>
        public static void Main(string[] args)
        {
            try
            {
                CheckConfigErrors();
                Init();
                if (IsServiceMode)
                {
                    Listening = false;
                    ServiceBase.Run(new Service());
                }
                else
                {
                    Listening = true;
                    ThreadProc();
                }
            }
            catch (Exception ex)
            {
                if (IsServiceMode)
                {
                    string errorMessage = "Could not start service\n     " + ex.Message;
                    logger.LogError(errorMessage, null);
                }
                else
                {
                    Console.WriteLine(ex.Message);
                    Console.WriteLine("\nPress any key...");
                    Console.ReadKey();
                }
            }
        }

        static void Init()
        {
            var exePath = Path.GetDirectoryName(new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath);
            string logPath = Path.Combine(Directory.GetParent(exePath).FullName, @"App_Data\WebDav\Logs");
            logger.LogFile = Path.Combine(logPath, "WebDAVlog.txt");
            logger.IsDebugEnabled = debugLoggingEnabled;

            engine = new DavEngine
            {
                Logger = logger,
                OutputXmlFormatting = logger.IsDebugEnabled ? Formatting.Indented : Formatting.None,
            };
            engine.License = File.ReadAllText(Path.Combine(exePath, "License.lic"));
            // set custom handler to process GET requests to folders and display info about how to connect to server
            MyCustomGetHandler handler = new MyCustomGetHandler();
            handler.OriginalHandler = engine.RegisterMethodHandler("GET", handler);
        }

        public static void ThreadProc()
        {
            string uriPrefix = ConfigurationManager.AppSettings["ListenerPrefix"];
            HttpListener listener = new HttpListener();
            listener.Prefixes.Add(uriPrefix);
            listener.AuthenticationSchemes = AuthenticationSchemes.Anonymous;

            listener.IgnoreWriteExceptions = true;
#if DEBUG
            if (!IsServiceMode)
#endif
#if DEBUG
            // For the sake of the development convenience, this code opens default web browser
            // with this server url when project is started in the debug mode as a console app.
            System.Diagnostics.Process.Start(uriPrefix.Replace("+", "localhost"), null);
#endif

            listener.Start();

            Console.WriteLine("Start listening on " + uriPrefix);
            Console.WriteLine("Press Control-C to stop listener...");

            while (Listening)
            {
                IAsyncResult result = listener.BeginGetContext(ListenerCallback, listener);
                result.AsyncWaitHandle.WaitOne();
            }
        }

        public static void ListenerCallback(IAsyncResult result)
        {
            HttpListener listener = (HttpListener)result.AsyncState;
            HttpListenerContext context = listener.EndGetContext(result);
            MacOsXPreprocessor.Process(context.Request); // fixes headers for Mac OS X v10.5.3 or later

            context.Response.SendChunked = false;
            using (var sqlDavContext = new DavContext(context, listener.Prefixes, context.User))
            {
                engine.Run(sqlDavContext);
            }

            CloseResponse(context);
        }

        private static void CloseResponse(HttpListenerContext context)
        {
            try
            {
                context.Response.Close();
            }
            catch
            {
                // client closed connection before the content was sent
            }
        }

        /// <summary>
        /// Checks configuration errors.
        /// </summary>
        private static void CheckConfigErrors()
        {
            string connStr = ConfigurationManager.ConnectionStrings["WebDAV"].ConnectionString;
            if (string.IsNullOrEmpty(connStr))
            {
                throw new Exception("SqlConnectionString section is missing or invalid!");
            }

            var exePath = Path.GetDirectoryName(new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath);
            var dataPath = Path.Combine(Directory.GetParent(exePath).FullName, "App_Data");
            connStr = connStr.Replace("|DataDirectory|", dataPath);
            // test database connection
            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();
            }

            string uriPrefix = ConfigurationManager.AppSettings["ListenerPrefix"];
            if (string.IsNullOrEmpty(uriPrefix))
            {
                throw new Exception("ListenerPrefix section is missing or invalid!");
            }
        }
    }
}
