//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (!Params.ACL)  return string.Empty; #>
//!<# if (Params.Storage != StorageType.Ntfs)  return string.Empty; #>
using System;
using System.Collections.Generic;
using System.DirectoryServices.AccountManagement;
using System.Linq;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;

namespace ITHit.WebDav.Server.SampleModel.__NtfsStorage__.Acl
{
    /// <summary>
    /// Logical folder which contains windows groups and is located
    /// at '/acl/groups' path.
    /// </summary>
    public class GroupFolder : LogicalFolder, IPrincipalFolder
    {
        /// <summary>
        /// Path to folder which contains groups.
        /// </summary>
        public const string PREFIX = "acl/groups";

        /// <summary>
        /// Initializes a new instance of the GroupFolder class.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/>.</param>
        public GroupFolder(DavContext context) : base(context, "groups", PATH)
        {
        }

        /// <summary>
        /// Gets path of folder which contains groups.
        /// </summary>
        public static string PATH
        {
            get { return PREFIX + "/"; }
        }

        /// <summary>
        /// Retrieves list of windows groups.
        /// </summary>
        /// <param name="properties">Properties which will be requested from the item by the engine later.</param>
        /// <returns>Enumerable with groups.</returns>
        public override IEnumerable<IHierarchyItem> GetChildren(IList<PropertyName> properties)
        {
            return Context.PrincipalOperation<IEnumerable<IHierarchyItem>>(getGroups);
        }

        /// <summary>
        /// Required by interface. However we don't allow creating folders inside this folder.
        /// </summary>
        /// <param name="name">New folder name.</param>
        /// <returns>New folder.</returns>
        public IPrincipalFolder CreateFolder(string name)
        {
            //Creating folders inside this folder is not supported.
            throw new DavException("Creating folders is not implemented", DavStatus.NOT_IMPLEMENTED);
        }

        /// <summary>
        /// Creates group with specified name.
        /// </summary>
        /// <param name="name">Group name.</param>
        /// <returns>Newly created group.</returns>
        public IPrincipal CreatePrincipal(string name)
        {
            if (!PrincipalBase.IsValidUserName(name))
            {
                throw new DavException("Group name contains invalid characters", DavStatus.FORBIDDEN);
            }

            var groupPrincipal = new GroupPrincipal(Context.GetPrincipalContext());

            groupPrincipal.Name = name;

            groupPrincipal.Save();

            return new Group(groupPrincipal, Context);
        }

        /// <summary>
        /// Finds groups which have matching property values.
        /// </summary>
        /// <param name="propValues">Property values which group must have.</param>
        /// <param name="props">Properties that will be retrieved later for the group found.</param>
        /// <returns>Groups which have matching property.</returns>
        public IEnumerable<IPrincipal> FindPrincipalsByPropertyValues(
            IList<PropertyValue> propValues,
            IList<PropertyName> props)
        {
            var group = new GroupPrincipal(Context.GetPrincipalContext());
            group.Name = "*";
            foreach (PropertyValue v in propValues)
            {
                if (v.QualifiedName == PropertyName.DISPLAYNAME)
                {
                    group.Name = "*" + v.Value + "*";
                }
            }

            var searcher = new PrincipalSearcher(group);
            return searcher.FindAll().Select(u => new Group((GroupPrincipal)u, Context)).Cast<IPrincipal>();
        }

        /// <summary>
        /// Returns properties which can be used in <see cref=" FindPrincipalsByPropertyValues"/> method.
        /// </summary>
        /// <returns>List of property description.</returns>
        public IEnumerable<PropertyDescription> GetPrincipalSearcheableProperties()
        {
            yield return new PropertyDescription
                             {
                                 Name = PropertyName.DISPLAYNAME,
                                 Description = "Principal name",
                                 Lang = "en"
                             };
        }

        /// <summary>
        /// Returns all groups current user is member of.
        /// </summary>
        /// <param name="props">Properties that will be asked later from the groups returned.</param>
        /// <returns>Enumerable with groups.</returns>
        public IEnumerable<IPrincipal> GetMatchingPrincipals(IList<PropertyName> props)
        {
            return User.FromName(Context.User.Name, Context).GetGroupMembership();
        }

        /// <summary>
        /// Retrieves all groups in computer/domain.
        /// </summary>
        /// <returns>Enumerable with groups.</returns>
        private IEnumerable<IHierarchyItem> getGroups()
        {
            var insGroupPrincipal = new GroupPrincipal(Context.GetPrincipalContext());

            insGroupPrincipal.Name = "*";
            var insPrincipalSearcher = new PrincipalSearcher(insGroupPrincipal);

            var r = insPrincipalSearcher.FindAll();

            return r.Select(g => new Group((GroupPrincipal)g, Context)).Cast<IHierarchyItem>().ToList();
        }
    }
}
